#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright 2014-2015 Michael Helmling
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation

from __future__ import division, print_function, unicode_literals
import argparse, io

from bibtexvcs.database import Database


def export(args):
    if args.template:
        with io.open(args.template, 'r', encoding='UTF-8') as templateFile:
            templateString = templateFile.read()
    else:
        templateString = None
    output = args.db.export(templateString=templateString, docDir=args.docs)
    if args.output == '-':
        print(output)
    else:
        with io.open(args.output, 'wt', encoding='UTF-8') as outfile:
            outfile.write(output)


def check(args):
    from bibtexvcs import checks
    errors, warnings = checks.performDatabaseCheck(args.db)
    for err in errors:
        print('FAIL: {}'.format(err))
    for warn in warnings:
        print('WARN: {}'.format(warn))


def script():
    """Command-line script that allows to export a database and run checks."""
    desc = ('Command-line interface to the BibTeX VCS package. Can be used to run the GUI, run '
            'JabRef configured for a specified BibTeX VCS database, export a database using '
            'templates (e.g. HTML output), or run database sanity checks.')
    parser = argparse.ArgumentParser(description=desc)
    parser.add_argument(
        '-d', '--database', metavar='DB',
        help='specify database root directory. If left out, the default database is used'
    )

    parser.add_argument('mode', choices=('gui', 'jabref', 'export', 'check'),
                        help='choose mode of operation (default: gui)',
                        default='gui',
                        nargs='?')

    exportGroup = parser.add_argument_group('exporting options (only in "export" mode)')
    exportGroup.add_argument('--template', help='template file')
    exportGroup.add_argument('--docs', help='documents root path')
    exportGroup.add_argument('output', nargs='?', help='output file')

    args = parser.parse_args()
    if args.mode == 'gui':
        import bibtexvcs.gui
        bibtexvcs.gui.run(args.database)
    else:
        # load database. We don't load it before starting the GUI because the GUI will display
        # a progress bar while loading the database by itself.
        if args.database:
            args.db = Database(args.database)
        else:
            args.db = Database.getDefault()
        if args.mode == 'export':
            export(args)
        elif args.mode == 'jabref':
            args.db.runJabref()
        elif args.mode == 'check':
            check(args)

if __name__ == '__main__':
    script()



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright 2014-2015 Michael Helmling
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation

from __future__ import division, print_function, unicode_literals
import unittest
from os.path import join, split

from bibtexvcs import database
from . import datadir

class TestDatabaseConfig(unittest.TestCase):

    def setUp(self):
        self.db = database.Database(join(datadir(), 'sampleDB'))

    def testConfValues(self):
        self.assertEqual(self.db.name, 'Example literature database managed by the bibtex VCS package')
        directory = split(self.db.directory)
        self.assertEqual(directory[-1], 'sampleDB')
        self.assertEqual(self.db.journalsName, 'journals.txt')
        self.assertEqual(self.db.bibfileName, 'sample.bib')





from os.path import abspath, join, dirname
from contextlib import contextmanager
import tempfile, shutil

from bibtexvcs import database, vcs


def datadir():
    return join(abspath(dirname(__file__)), 'data')



@contextmanager    
def tmpDatabase():
    tmpdir = tempfile.mkdtemp()
    newDir = join(tmpdir, 'btvcs')
    shutil.copytree(join(datadir(), 'sampleDB'), newDir)
    try:
        yield database.Database(newDir)
    finally:
        shutil.rmtree(tmpdir)

@contextmanager
def tmpClonedDatabase(source):
    tmpdir = tempfile.mkdtemp()
    newDir = join(tmpdir, 'btvcs')
    try:
        yield vcs.VCSInterface.getClonedDatabase(source, newDir, 'mercurial')
    finally:
        shutil.rmtree(tmpdir)


#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright 2014-2015 Michael Helmling
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation

from __future__ import division, print_function, unicode_literals
import os, subprocess, sys

from bibtexvcs import config


class Login:
    """Data class representing login information for a remote repository.
    """
    def __init__(self, username=None, password=None):
        self.username = username
        self.password = password


class MergeConflict(Exception):
    """Raised when merge conflicts occur on a VCS update operation."""
    pass


class AuthError(Exception):
    """Raised on authentication errors when accessing remote repositories."""
    pass


class VCSNotFoundError(Exception):
    """Raised when the necessary VCS binary (e.g. "hg") is not installed."""


class VCSInterface:
    """Interface to the version control system (VCS) of a :mod:`bibtexvcs` database.
    """

    __typeMap = {}

    def __init__(self, database):
        self.database = database
        storedLogin = Login(*config.getLogin(database))
        self.login = storedLogin

    @property
    def root(self):
        """Returns the root directory of this repository.
        """
        return self.database.directory

    def storeLogin(self, login):
        """Permanently store login information for this repository.
        """
        config.storeLogin(self.database, self.login.username, self.login.password)

    def add(self, path):
        """Adds a file or directory to the repository.
        """

    def hasLocalChanges(self):
        """Determines whether any local changes have been made to the repository.
        Returns
        -------
        bool
            Return value is ``True`` iff one of the following requirements are met:

            - any versioned file has been modified,
            - any versioned file in the documents folder has been deleted, or
            - any unversioned file has been placed in the documents folder.
        """
        raise NotImplementedError()

    def update(self):
        """Checks if there are updates in the remote repository. If so, tries to merge
        them and reloads the database.

        Returns
        -------
        bool
            Flag indicating if something was updated or not.
        Raises
        ------
        MergeConflict
            If update results in a merge that cannot be handled automatically.
        """
        raise NotImplementedError()

    def commit(self, commitMessage=None):
        """Commit local changes and push to remote, if remote is configured.

        Returns
        ----------
        bool
            Indicator being ``True`` iff there actually were any changes.
        """
        raise NotImplementedError()

    def revision(self):
        """Returns a tuple containing the current revision and its date."""
        raise NotImplementedError()


    @staticmethod
    def vcsTypeNames():
        """Returns the names of all registered VCS types."""
        return VCSInterface.__typeMap.keys()

    @staticmethod
    def registerVCSType(vcsTypeName, vcsTypeClass):
        """Register a new VCS type."""
        VCSInterface.__typeMap[vcsTypeName] = vcsTypeClass

    @staticmethod
    def getImplementation(vcsTypeName):
        if vcsTypeName == 'local':
            return None
        try:
            return VCSInterface.__typeMap[vcsTypeName]
        except KeyError:
            raise KeyError('No VCS implementation of type "{}" available'.format(vcsTypeName))

    @staticmethod
    def get(database):
        return VCSInterface.getImplementation(database.vcsType)(database)

    @classmethod
    def clone(cls, url, target, login=None):
        """Clones a remote repository.

        Parameters
        ----------
        url : str
            The remote repository URL.
        target : str
            The local target directory.
        login : Login
            Login information.
        """
        raise NotImplementedError()

    @staticmethod
    def getClonedDatabase(url, target, vcsType, login=None, storeLogin=False):
        """Clone and return a database from a remote location specified by `url`.

        Parameters
        ----------
        url : str
            Source URL of the repository.
        target : str
            Local target directory.
        vcsType
            VCS system to use.
        login : Login
            Login information (optional).
        storeLogin : bool (optional)
            Store login on successful authentication.
        """
        vcsCls = VCSInterface.getImplementation(vcsType)
        if login is None:
            login = Login()
        vcsCls.clone(url, target, login)
        from bibtexvcs.database import Database
        database = Database(target)
        # copy login information
        database.vcs.login = login
        if storeLogin:
            database.vcs.storeLogin()
        return database


class MercurialInterface(VCSInterface):
    """Interface to the `Mercurial <http://mercurial.selenic.com>`_ version control system.
    """

    cmdline = ['hg', '--noninteractive', '--config', 'auth.x.prefix=*']

    def __init__(self, *args, **kwargs):
        VCSInterface.__init__(self, *args, **kwargs)
        try:
            self.hasRemote = len(self.callHg('showconfig', 'paths.default')) > 0
        except subprocess.CalledProcessError:
            # some hg versions appear to return exit status 1 if req'ed config noit set
            self.hasRemote = False
        self.hgid = self.callHg('id', '-i')

    def callHg(self, *args):
        """Calls the ``hg`` script in the database directory with the given arguments.
        """
        return MercurialInterface._callHg(*args, login=self.login, cwd=self.root)

    @classmethod
    def _callHg(cls, *args, **kwargs):
        cmdline = cls.cmdline[:]
        kwargs = kwargs.copy()
        if kwargs.get('login'):
            login = kwargs['login']
            if login.username:
                cmdline += ['--config', 'auth.x.username={}'.format(login.username)]
            if login.password:
                cmdline += ['--config', 'auth.x.password={}'.format(login.password)]
        if 'login' in kwargs:
            del kwargs['login']
        env = os.environ.copy()
        env['LANG'] = 'C'
        env['LANGUAGE'] = 'C'  # on Ubuntu 12.04, LANG=C does not convince hg to use english output
        try:
            return subprocess.check_output(cmdline + list(args), env=env,
                                           stderr=subprocess.STDOUT, **kwargs)
        except subprocess.CalledProcessError as e:
            output = e.output.decode(errors='replace')
            if "authorization required" in output:
                raise AuthError('Authorization required for the mercurial repository.')
            if "authorization failed" in output:
                raise AuthError('Authorization for the mercurial repository failed.')
            if 'conflicts during merge' in output:
                raise MergeConflict('Conflict arised when merging remote and local changes!\n'
                                    'You need to fix this issue by hand. Error message:\n{}'
                                    .format(e.output))
            if 'unresolved merge conflicts' in output:
                raise MergeConflict('There are unresolved merge conflicts in your repository.\n'
                                    'You have to fix them manually before proceeding to use this '
                                    'tool.')
            print(output)
            raise e
        except OSError as e:
            raise VCSNotFoundError('Could not run "hg". Please install mercurial')

    def add(self, path):
        self.callHg('add', path)

    def hasLocalChanges(self):
        if len(self.callHg('status', '--modified', '--added')) > 0:
            return True
        # by this we ignore unversioned files in the base folder
        return len(self.callHg('status', '--deleted', '--unknown', self.database.documents)) > 0

    def update(self):
        if self.hasRemote:
            self.callHg('pull')
        # force internal merge algorithm with "-t merge" to prevent GUI tools opening
        self.callHg('update', '-t', ':merge')
        newId = self.callHg('id', '-i')
        if newId != self.hgid:
            self.hgid = newId
            self.database.reload()
            return True
        return False

    def commit(self, commitMessage=None):
        if not self.hasLocalChanges():
            return False
        hgOutput = self.callHg('status', '--deleted', '--no-status', self.database.documents)
        deletedDocs = hgOutput.decode(sys.getfilesystemencoding()).splitlines()
        if len(deletedDocs) > 0:
            self.callHg('remove', *deletedDocs)
        hgOutput = self.callHg('status', '--unknown', '--no-status', self.database.documents)
        newDocs = hgOutput.decode().splitlines()
        if len(newDocs) > 0:
            self.callHg('add', *newDocs)
        # TODO: sophisticated analysis of diff to current head
        self.callHg('commit', '--message', commitMessage or 'Auto-Commit by BibTeX VCS')
        if self.hasRemote:
            self.callHg('push')
        self.database.reload()
        self.hgid = self.callHg('id', '-i')
        return True

    def revision(self):
        return self.callHg('log', '-l', '1', '--template', '{rev}\n{date|isodate}'
                           ).decode().splitlines()

    @classmethod
    def clone(cls, url, target, login=None):
        if login is None:
            login = Login()
        cls._callHg('clone', url, target, login=login)


VCSInterface.registerVCSType('mercurial', MercurialInterface)


#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright 2014-2015 Michael Helmling
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation
from __future__ import division, print_function, unicode_literals
import io, re, os, sys

from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
# From PyPA's sampleproject (https://github.com/pypa/sampleproject).
with io.open(os.path.join(here, 'bibtexvcs', '__init__.py'), 'r', encoding='UTF-8') as f:
    version_file = f.read()
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        version = version_match.group(1)
    else:
        raise RuntimeError("Unable to find version string.")

with io.open(os.path.join(here, 'README.rst'), encoding='UTF-8') as f:
    long_description = f.read()

requires = ['pyparsing']
if sys.version_info.major == 2:
    # depend on backported packages
    requires.append('futures')
    requires.append('configparser')


setup(
    name='bibtexvcs',
    version=version,
    description="a Python package for managing a BibTeX database and related documents",
    long_description=long_description,
    url='http://github.com/supermihi/bibtevxcs',
    author='Michael Helmling',
    author_email='michaelhelmling@posteo.de',
    classifiers=[
      'Development Status :: 4 - Beta',
      'Intended Audience :: Science/Research',
      'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
      'Operating System :: OS Independent',
      'Programming Language :: Python :: 2.7',
      'Programming Language :: Python :: 3',
      'Topic :: Database :: Front-Ends',
    ],
    license='GPL3',
    keywords='bibliography bibtex jabref',
    packages=find_packages(),
    install_requires=requires,
    entry_points=dict(gui_scripts=['btvcs = bibtexvcs.script:script']),
    include_package_data=True,
    test_suite='test'
)



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright 2014-2015 Michael Helmling
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation

"""
The :mod:`checks` module defines a framework for defining and running automated *checks* on a
bibtexvcs
database.

Checks define certain constraints on the database related to consistency or other regulations.
For instance, a default check ensures that every macro that is used in the BibTeX file is defined
either in the bib file or in the list of journals.

Custom checks can be added per database by creating a module ``checks.py`` in the main database
directory. In that file, each check should be implemented as a function decorated with the
:func:`databaseCheck` decorator that takes the check's description as an argument. The check
function needs to act as a generator that yields :class:`CheckFailed` or :class:`CheckWarning`
instances in case of failures or warnings. An example file might look like this (contains a check
ensuring that every *journal* or *inproceedings* entry is a macro::

    from bibtexvcs.checks import CheckFailed, CheckWarning, databaseCheck
    from bibtexvcs.bibfile import MacroReference

    @databaseCheck('only macros in journals')
    def checkMacrosInJournals(database):
        for entry in database.bibfile.values():
            for field in ('inproceedings', 'journal'):
                if field in entry and not isinstance(entry[field], MacroReference):
                    yield CheckFailed('Entry "{}" has a non-macro {} field "{}"'
                                      .format(entry.citekey, field, entry[field]))
"""


from __future__ import division, print_function, unicode_literals
import os.path
import sys
import inspect
import imp
from bibtexvcs.bibfile import MacroReference, MONTHS


def performDatabaseCheck(database, exclude=[]):
    me = sys.modules[__name__]
    checks = {fun.checkName: fun for fname, fun in inspect.getmembers(me, inspect.isfunction)
                                 if hasattr(fun, 'checkName')}
    if os.path.exists(database.checksPath):
        localChecks = imp.load_source('checks', database.checksPath)
        for fname, fun in inspect.getmembers(localChecks, inspect.isfunction):
            if hasattr(fun, 'checkName'):
                checks[fun.checkName] = fun
    errors = []
    warnings = []

    for checkName, check in checks.items():
        if checkName not in exclude:
            for ans in check(database):
                if isinstance(ans, CheckFailed):
                    errors.append(ans)
                else:
                    warnings.append(ans)
    return errors, warnings


class CheckFailed(Exception):
    pass


class CheckWarning(Warning):
    pass


def databaseCheck(name):
    """Function decorator for database checks.

    :param name: a (unique) description text for the check.
    """
    def wrap(func):
        func.checkName = name
        return func
    return wrap


@databaseCheck('macro references')
def checkMacros(database):
    """Check if all macros referenced in the database exist in its journals file."""
    bib = database.bibfile
    for entry in bib.values():
        for field, value in entry.items():
            if isinstance(value, MacroReference):
                if value.name in bib.macroDefinitions or value.name in database.journals:
                    continue
                if value.name in MONTHS:
                    continue
                yield CheckFailed("The macro '{m}' used for field '{f}' in bibtex "
                                  "entry '{e}' is defined neither in the database nor "
                                  "in the journals file."
                                  .format(f=field, m=value.name, e=entry.citekey))


@databaseCheck('file links')
def checkFileLinks(database):
    """Check that the files linked to in the database match those existing in the `documents`
    directory. Additionally, check that all documents are contained in the `documents` directory.
    """

    dbFilesSet = set(database.referencedDocuments())
    fsFilesSet = set(database.existingDocuments())

    if len(dbFilesSet - fsFilesSet) > 0:
        yield CheckFailed("The following file(s) are referenced in the bibtex file but do not "
                          "exist in the documents directory:\n{}"
                          .format("\n".join(dbFilesSet - fsFilesSet)))
    if len(fsFilesSet - dbFilesSet) > 0:
        yield CheckFailed("The following file(s) in the documents directory are not referenced by "
                          "any entry in the bibtex file:\n{}"
                          .format("\n".join(fsFilesSet - dbFilesSet)))


@databaseCheck('ASCII filenames')
def checkASCIIFilenames(database):
    """Check that all file names are ASCII. This is sensible because non-ASCII file names lead
    to problems with most VCS systems.
    """
    for filename in database.existingDocuments():
        try:
            filename.encode('ascii')
        except UnicodeEncodeError:
            yield CheckFailed('The file name "{}" contains non-ASCII characters.'.format(filename))


@databaseCheck('month macros')
def checkMonthMacros(database):
    """Checks that the ``month`` field only contains (proper) macros."""
    for entry in database.bibfile.values():
        if 'month' not in entry:
            continue
        month = entry['month']
        if isinstance(month, str):
            yield CheckFailed("Month field in entry '{e}' contains the string '{s}' instead "
                              "of a month macro.".format(e=entry.citekey, s=month))
        elif isinstance(month, MacroReference):
            if month.name not in MONTHS:
                yield CheckFailed("Invalid month macro '{}' used in entry '{}'"
                                 .format(month.name, entry.citekey))
        else:
            #  must be a list of macros
            if len(month) % 2 != 1:
                yield CheckFailed("Invalid month definition '{}' in '{}': Must be either a single "
                                 "month macro or of the format 'mar / apr'."
                                 .format(month, entry.citekey))
                continue
            separators = [ month[i] for i in range(1, len(month), 2) ]
            macros = [ month[i] for i in range(0, len(month), 2) ]
            for separator in separators:
                if separator.strip() != '/':
                    yield CheckFailed("Invalid month definition '{}' in '{}': Expected '/' but got "
                                      "'{}".format(month, entry.citekey, separator.strip()))
            for macro in macros:
                if not isinstance(macro, MacroReference) or macro.name not in MONTHS:
                    yield CheckFailed("Invalid month definition '{}' in '{}'"
                                      .format(month, entry.citekey))


@databaseCheck('jabref file directory')
def checkJabrefFileDirectory(database):
    identifier = 'jabref-meta: fileDirectory:'
    for comment in database.bibfile.comments:
        if comment.comment.startswith(identifier):
            if comment.comment[len(identifier):] != database.documents + ';':
                yield CheckFailed('JabRef fileDirectory field does not coincide with '
                                  'the configured one.')


@databaseCheck('required BibTeX fields')
def checkRequiredFields(database):
    """Checks that all required fields exist for each entry."""
    requiredFields = {
        'article'      : ('author', 'title', 'journal', 'year'),
        'book'         : (('author', 'editor'), 'title', 'publisher', 'year'),
        'booklet'      : ('title',),
        'incollection' : ('author', 'title', 'booktitle', 'publisher', 'year'),
        'inproceedings': ('author', 'title', 'booktitle', 'year'),
        'mastersthesis': ('author', 'title', 'school', 'year'),
        'phdthesis'    : ('author', 'title', 'school', 'year'),
        'misc'         : (),
        'techreport'   : ('author', 'title', 'institution', 'year'),
        'unpublished'  : ('author', 'title', 'note'),
        'online'       : (('author', 'editor'), 'title', 'year', 'url')
    }
    for entry in database.bibfile.values():
        if entry.entrytype not in requiredFields:
            yield CheckWarning('Entry "{}": Required fields for type "{}" unknown'
                               .format(entry.citekey, entry.entrytype))
        else:
            for req in requiredFields[entry.entrytype]:
                if isinstance(req, tuple):
                    if not any(subReq in entry for subReq in req):
                        yield CheckFailed('Entry "{}" of type "{}" requires one of the fields: {}'
                                          .format(entry.citekey, entry.entrytype, ', '.join(req)))
                elif req not in entry:
                    yield CheckFailed('Entry "{}" of type "{}" requires field "{}"'
                                      .format(entry.citekey, entry.entrytype, req))


@databaseCheck('entry owners')
def checkOwnerExists(database):
    for entry in database.bibfile.values():
        if 'owner' not in entry:
            yield CheckFailed('Entry "{}" has no owner.'.format(entry.citekey))


@databaseCheck('marked entries')
def checkNoMarkedEntry(database):
    for entry in database.bibfile.values():
        if '__markedentry' in entry:
            yield CheckFailed('Entry "{}" is marked in jabref:\n{}'
                              .format(entry.citekey, entry['__markedentry']))


#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright 2014-2015 Michael Helmling
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation

from __future__ import division, print_function, unicode_literals
from collections import OrderedDict
import io

"""This module contains classes for an object-oriented representation of the .bib file.

Most of it is generic for BibTeX, i.e. not special to the literature package.
"""


class BibFile(OrderedDict):
    """Object-oriented encapsulation of a BibTeX database.

    :param filename: Path of the ``.bib`` file to read.
    :type filename: str

    :param bibstring: BibTeX database as string (as an alternative to ``filename``)
    :type bibstring: str

    .. attribute:: macroDefinitions

        Dictionary of :class:`MacroReference` objects defined in this bib file.
    """

    def __init__(self, filename=None, bibstring=None):
        super(BibFile, self).__init__()
        self.filename = filename
        from . import parser
        if filename:
            with io.open(filename, "rt", encoding='UTF-8') as bibFile:
                bibstring = bibFile.read()
        bibParsed = parser.bibfile.parseString(bibstring, parseAll=True)
        self.comments = []
        self.macroDefinitions = OrderedDict()
        for item in bibParsed:
            if isinstance(item, Entry):
                self[item.citekey] = item
            elif isinstance(item, Comment):
                self.comments.append(item)
            elif isinstance(item, MacroDefinition):
                self.macroDefinitions[item.key] = item
            elif isinstance(item, Preamble):
                self.preamble = item
            else:
                raise ValueError('Unknown item parsed: {}'.format(item))


class DatabaseFormatError(Exception):
    """Raised if the BibTeX database file is malformed."""
    pass


class DatabaseElement:
    """Base class for database elements (Entries, Comments, Macros)."""

    @classmethod
    def fromParseResult(cls, toks):
        """Creates the database element from the result of the parser.

        This method is meant to be used in :mod:`.parser` as parseAction on the corresponding element.

        @param toks: the token list
        """
        raise NotImplementedError()


class MacroDefinition(DatabaseElement):
    """Represents a macro definition of the form ``@STRING{bla={blub}}``.

    The members :attr:`key` and :attr:`value` hold the corresponding elements of the definition.

    Attributes
    ----------
    key : str
        The key by which the defined macro is accessed.

    value : str
        The substitution string of the macro definition.
    """

    def __init__(self, macro, definition):
        self.key = macro
        self.value = definition

    def __str__(self):
        return "MacroDefinition({}={})".format(self.key, self.value)

    @classmethod
    def fromParseResult(cls, toks):
        macro = toks["macro"]
        definition = toks["definition"]
        return [cls(macro, definition)]


class MacroReference(DatabaseElement):
    """ Class to encapsulate undefined macro references."""
    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return 'MacroReference("{}")'.format(self.name)

    __str__ = __repr__

    def __eq__(self, other):
        return self.name == other.name

    def __ne__(self, other):
        return self.name != other.name

    @classmethod
    def fromParseResult(cls, toks):
        return [cls(toks[0])]


class Entry(DatabaseElement, OrderedDict):
    """A BibTeX entry.
    """

    def __init__(self, entrytype, citekey, fields, src):
        OrderedDict.__init__(self)
        self.bibsrc = src
        for key, val in fields.items():
            self[key] = val
        self.citekey = citekey
        self.entrytype = entrytype

    @classmethod
    def fromParseResult(cls, toks):
        def formatValue(value):
            """Joins together (nested) bracketed strings."""
            if isinstance(value, str) or not hasattr(value, '__iter__'):
                return value
            parts = [formatValue(v) for v in value]
            if all(isinstance(v, str) for v in parts):
                return "".join(parts)
            return parts
        bibsrc = toks[-1]
        fields = {}
        for key, val in toks.items():
            if key == "entry type":
                entrytype = val
            elif key == "cite key":
                citekey = val
            else:
                fields[key] = formatValue(val)
        return [Entry(entrytype=entrytype, citekey=citekey, fields=fields, src=bibsrc)]

    def filename(self):
        """Returns the filename referenced in the BibTeX ``file`` field in `JabRef`_'s format.

        The format of the ``file`` field is::

            :filename:

        where ``filename`` is relative to the documents directory.
        If a ``file`` field is present but does not match this format,
        a :class:`DatabaseFormatError` will be raised.

        :returns: The filename or `None` if there is no ``file`` field in the entry.

        """
        if "file" in self:
            fname = self["file"]
            try:
                return fname[1:].rsplit(":", 1)[0].replace("\;", ";")
            except IndexError:
                raise DatabaseFormatError('Wrong file URL format in entry "{}": '
                                          '{1}'.format(self.citekey, fname))

    def doiURL(self):
        """Returns the DOI URL if a "doi" field is present or None otherwise."""
        if "doi" in self:
            return "http://dx.doi.org/" + self["doi"]

    def lastNames(self, field='author', maxNames=3):
        if field not in self:
            return None
        names = self[field]
        if isinstance(names, str) or isinstance(names, Name):
            return str(names)
        return ", ".join(name.lastName() for name in names[:maxNames]) \
            + (" et al."if len(names) > maxNames else "")

    def __str__(self):
        return "{}({}) by {}".format(self.entrytype, self.citekey, self.get("author"))


class Comment(DatabaseElement):
    """Represents a comment @COMMENT{<text>}"""
    def __init__(self, comment):
        self.comment = comment

    @classmethod
    def fromParseResult(cls, toks):
        return [cls(toks["comment"])]

    def __str__(self):
        return "{}({})".format(self.__class__.__name__, self.comment)


class ImplicitComment(Comment):
    """Implicit comment in the bib file (everything before the first "@" symbol)."""
    pass


class Preamble(DatabaseElement):
    def __init__(self, contents):
        self.contents = contents

    @classmethod
    def fromParseResult(cls, toks):
        return [cls(toks["preamble"])]


class Name:
    def __init__(self, last, nobility=None, first=None, suffix=None):
        self.first = first
        self.nobility = nobility
        self.last = last
        self.suffix = suffix

    def __str__(self):
        return self.last

    def lastName(self):
        """Return a formatted version of the last name, including nobility and suffix
        (if appropriate).
        """
        return ' '.join((part for part in (self.nobility, self.last, self.suffix) if part))


MONTHS = dict((month[:3].lower(), MacroDefinition(month[:3].lower(), month)) for month in
          ("January", "February", "March", "April", "May", "June", "July", "August",
           "September", "October", "November", "December"))
